@@ -41,31 +41,35 @@
 #define HANDLE_KEY(key, fn) case key: fn(); break;
 
 int count, reg;
-
 /*
- * Store the events and appointments for the selected day and reset the
- * appointment highlight pointer if a new day was selected.
+ * Store events and appointments for a range of days in the day vector -
+ * beginning with the selected day - and load them into the APP listbox. If no
+ * day-change occurs, reset the selected APP item and with it the selected day,
+ * thereby storing and loading the same range of days.
  */
 static void do_storage(int day_changed)
 {
-	struct day_item *day = ui_day_selitem();
-	union aptev_ptr item;
+	/*
+	 * Save the selected item before rebuilding the day vector -
+	 * unless already set.
+	 */
+	if (!day_check_sel_data())
+		day_set_sel_data(ui_day_get_sel());
 
-	if (day) {
-		/*
-		 * day_process_storage() rebuilds the vector of day items, so
-		 * we need to save the reference to the actual item here.
-		 */
-		item = day->item;
-	}
+	if (!day_changed)
+		ui_day_sel_reset();
 
-	day_process_storage(ui_calendar_get_slctd_day(), day_changed);
+	/* The day_items vector. */
+	day_store_items(get_slctd_day(), 1, day_get_days());
+	/* The APP listbox. */
 	ui_day_load_items();
 
 	if (day_changed)
 		ui_day_sel_reset();
-	else if (day)
-		ui_day_set_selitem_by_aptev_ptr(item);
+	else
+		ui_day_find_sel();
+
+	day_set_sel_data(&empty_day);
 }
 
 static inline void key_generic_change_view(void)
@@ -77,7 +81,7 @@ static inline void key_generic_change_view(void)
 
 static inline void key_generic_other_cmd(void)
 {
-	wins_other_status_page(wins_slctd());
+	wins_other_status_page();
 	wins_update(FLAG_STA);
 }
 
@@ -101,7 +105,7 @@ static inline void key_generic_goto_today(void)
 
 static inline void key_view_item(void)
 {
-	if (wins_slctd() == APP)
+	if (wins_slctd() == APP && !event_dummy(ui_day_get_sel()))
 		ui_day_popup_item();
 	else if (wins_slctd() == TOD)
 		ui_todo_popup_item();
@@ -120,6 +124,7 @@ static inline void key_generic_config_menu(void)
 static inline void key_generic_add_appt(void)
 {
 	ui_day_item_add();
+	do_storage(0);
 	wins_update(FLAG_CAL | FLAG_APP | FLAG_STA);
 }
 
@@ -135,6 +140,7 @@ static inline void key_add_item(void)
 	case APP:
 	case CAL:
 		ui_day_item_add();
+		do_storage(0);
 		wins_update(FLAG_CAL | FLAG_APP | FLAG_STA);
 		break;
 	case TOD:
@@ -148,7 +154,7 @@ static inline void key_add_item(void)
 
 static inline void key_edit_item(void)
 {
-	if (wins_slctd() == APP) {
+	if (wins_slctd() == APP && !event_dummy(ui_day_get_sel())) {
 		ui_day_item_edit();
 		do_storage(0);
 		wins_update(FLAG_CAL | FLAG_APP | FLAG_STA);
@@ -160,7 +166,7 @@ static inline void key_edit_item(void)
 
 static inline void key_del_item(void)
 {
-	if (wins_slctd() == APP) {
+	if (wins_slctd() == APP && !event_dummy(ui_day_get_sel())) {
 		ui_day_item_delete(reg);
 		do_storage(0);
 		wins_update(FLAG_CAL | FLAG_APP | FLAG_STA);
@@ -172,11 +178,8 @@ static inline void key_del_item(void)
 
 static inline void key_generic_copy(void)
 {
-	if (wins_slctd() == APP) {
+	if (wins_slctd() == APP && !event_dummy(ui_day_get_sel()))
 		ui_day_item_copy(reg);
-		do_storage(0);
-		wins_update(FLAG_CAL | FLAG_APP);
-	}
 }
 
 static inline void key_generic_paste(void)
@@ -190,15 +193,16 @@ static inline void key_generic_paste(void)
 
 static inline void key_repeat_item(void)
 {
-	if (wins_slctd() == APP)
+	if (wins_slctd() == APP && !event_dummy(ui_day_get_sel())) {
 		ui_day_item_repeat();
-	do_storage(0);
-	wins_update(FLAG_CAL | FLAG_APP | FLAG_STA);
+		do_storage(0);
+		wins_update(FLAG_CAL | FLAG_APP | FLAG_STA);
+	}
 }
 
 static inline void key_flag_item(void)
 {
-	if (wins_slctd() == APP) {
+	if (wins_slctd() == APP && !event_dummy(ui_day_get_sel())) {
 		ui_day_flag();
 		do_storage(0);
 		wins_update(FLAG_APP);
@@ -237,7 +241,7 @@ static inline void key_lower_priority(void)
 
 static inline void key_edit_note(void)
 {
-	if (wins_slctd() == APP) {
+	if (wins_slctd() == APP && !event_dummy(ui_day_get_sel())) {
 		ui_day_edit_note();
 		do_storage(0);
 	} else if (wins_slctd() == TOD) {
@@ -248,7 +252,7 @@ static inline void key_edit_note(void)
 
 static inline void key_view_note(void)
 {
-	if (wins_slctd() == APP)
+	if (wins_slctd() == APP && !event_dummy(ui_day_get_sel()))
 		ui_day_view_note();
 	else if (wins_slctd() == TOD)
 		ui_todo_view_note();
@@ -271,18 +275,72 @@ static inline void key_generic_help(void)
 
 static inline void key_generic_save(void)
 {
-	io_save_cal(IO_SAVE_DISPLAY_BAR);
-	wins_update(FLAG_STA);
+	char *msg = NULL;
+	int ret;
+
+	ret = io_save_cal(interactive);
+
+	if (ret == IO_SAVE_RELOAD) {
+		ui_todo_load_items();
+		ui_todo_sel_reset();
+		do_storage(0);
+		notify_check_next_app(1);
+		ui_calendar_monthly_view_cache_set_invalid();
+	}
+	wins_update(FLAG_ALL);
+	switch (ret) {
+	case IO_SAVE_CTINUE:
+		msg = _("Data were saved successfully");
+		break;
+	case IO_SAVE_RELOAD:
+		msg = _("Data were saved/reloaded successfully");
+		break;
+	case IO_SAVE_CANCEL:
+		msg = _("Save cancelled");
+		break;
+	case IO_SAVE_NOOP:
+		msg = _("Data were already saved");
+		break;
+	case IO_SAVE_ERROR:
+		EXIT(_("Cannot open data file"));
+	}
+	status_mesg(msg, "");
 }
 
 static inline void key_generic_reload(void)
 {
-	if (io_reload_data()) {
+	char *msg = NULL;
+	int ret;
+
+	ret = io_reload_data();
+	if (ret == IO_RELOAD_LOAD ||
+	    ret == IO_RELOAD_CTINUE ||
+	    ret == IO_RELOAD_MERGE) {
+		ui_todo_load_items();
+		ui_todo_sel_reset();
 		do_storage(0);
 		notify_check_next_app(1);
 		ui_calendar_monthly_view_cache_set_invalid();
-		wins_update(FLAG_ALL);
 	}
+	wins_update(FLAG_ALL);
+	switch (ret) {
+	case IO_RELOAD_LOAD:
+	case IO_RELOAD_CTINUE:
+		msg = _("Data were reloaded successfully");
+		break;
+	case IO_RELOAD_MERGE:
+		msg = _("Date were merged/reloaded successfully");
+		break;
+	case IO_RELOAD_CANCEL:
+		msg = _("Reload cancelled");
+		break;
+	case IO_RELOAD_NOOP:
+		msg = _("Data were already loaded");
+		break;
+	case IO_RELOAD_ERROR:
+		EXIT(_("Cannot open data file"));
+	}
+	status_mesg(msg, "");
 }
 
 static inline void key_generic_import(void)
@@ -356,8 +414,12 @@ static inline void key_move_up(void)
 	if (wins_slctd() == CAL) {
 		key_generic_prev_week();
 	} else if (wins_slctd() == APP) {
-		ui_day_sel_move(-1);
-		wins_update(FLAG_APP);
+		if (!ui_day_sel_move(-1)) {
+			ui_calendar_move(DAY_PREV, 1);
+			do_storage(1);
+			ui_day_sel_dayend();
+		}
+		wins_update(FLAG_APP | FLAG_CAL);
 	} else if (wins_slctd() == TOD) {
 		ui_todo_sel_move(-1);
 		wins_update(FLAG_TOD);
@@ -376,8 +438,12 @@ static inline void key_move_down(void)
 	if (wins_slctd() == CAL) {
 		key_generic_next_week();
 	} else if (wins_slctd() == APP) {
-		ui_day_sel_move(1);
-		wins_update(FLAG_APP);
+		if (!ui_day_sel_move(1)) {
+			ui_calendar_move(DAY_PREV, day_get_days() - 2);
+			do_storage(1);
+			ui_day_sel_daybegin(day_get_days() - 1);
+		}
+		wins_update(FLAG_APP | FLAG_CAL);
 	} else if (wins_slctd() == TOD) {
 		ui_todo_sel_move(1);
 		wins_update(FLAG_TOD);
@@ -454,34 +520,44 @@ static inline void key_generic_scroll_down(void)
 
 static inline void key_generic_quit(void)
 {
-	if (conf.auto_save)
-		io_save_cal(IO_SAVE_DISPLAY_BAR);
+	/* In read-only mode, quit unconditionally without saving. */
+	if (!read_only &&
+	    (conf.auto_save || (io_get_modified() &&
+	     status_ask_bool(_("There are unsaved changes. "
+	    "Should they be saved?")) == 1)))
+		if (io_save_cal(interactive) == IO_SAVE_CANCEL) {
+			/* Cancel quit as well. */
+			wins_update(FLAG_STA);
+			return;
+		}
 	if (conf.auto_gc)
 		note_gc();
 
 	if (conf.confirm_quit) {
-		if (status_ask_bool(_("Do you really want to quit?")) == 1) {
+		if (status_ask_bool(_("Do you really want to quit?")) == 1)
 			exit_calcurse(EXIT_SUCCESS);
-		} else {
-			wins_erase_status_bar();
+		else
 			wins_update(FLAG_STA);
-		}
 	} else {
 		exit_calcurse(EXIT_SUCCESS);
 	}
 }
 
+/*
+ * Safety exit.
+ * Auto_save is ignored, but modifications are checked for.
+ * Use of force(=!) will override configuration settings and --read-only option.
+ */
 static inline void key_generic_cmd(void)
 {
 	char cmd[BUFSIZ] = "";
 	char *cmd_name;
-	int valid = 0, force = 0;
+	int valid = 0, force = 0, ret;
 	char *error_msg;
 
-	status_mesg(_("Command:"), "");
+	status_mesg(_("Command: [ h(elp) | w(rite)(!) | q(uit)(!) | wq(!) ]"), "");
 	if (getstring(win[STA].p, cmd, BUFSIZ, 0, 1) != GETSTRING_VALID)
 		goto cleanup;
-
 	cmd_name = strtok(cmd, " ");
 	if (cmd_name[strlen(cmd_name) - 1] == '!') {
 		cmd_name[strlen(cmd_name) - 1] = '\0';
@@ -490,20 +566,35 @@ static inline void key_generic_cmd(void)
 
 	if (!strcmp(cmd_name, "write") || !strcmp(cmd_name, "w") ||
 	    !strcmp(cmd_name, "wq")) {
-		io_save_cal(IO_SAVE_DISPLAY_BAR);
+		if (force)
+			read_only = 0;
+		ret = io_save_cal(interactive);
+		/* Either the save was cancelled or read_only mode is on */
+		if (ret == IO_SAVE_CANCEL) {
+			if (read_only) {
+				status_mesg(_("Read-only mode - use w!"), "");
+				return;
+			} else if (!strcmp(cmd_name, "wq")) {
+				/* Cancel quit as well. */
+				goto cleanup;
+			}
+		}
 		valid = 1;
 	}
 	if (!strcmp(cmd_name, "quit") || !strcmp(cmd_name, "q") ||
 	    !strcmp(cmd_name, "wq")) {
+		if (!force && io_get_modified()) {
+			status_mesg(
+			    _("There are unsaved changes - use w or q!"), "");
+			return;
+		}
 		if (force || !conf.confirm_quit || status_ask_bool(
-				_("Do you really want to quit?")) == 1)
+		    _("Do you really want to quit?")) == 1)
 			exit_calcurse(EXIT_SUCCESS);
-		else
-			wins_erase_status_bar();
 		valid = 1;
 	}
 
-	if (!strcmp(cmd_name, "help")) {
+	if (!strcmp(cmd_name, "help") || !strcmp(cmd_name, "h")) {
 		char *topic = strtok(NULL, " ");
 
 		if (!display_help(topic)) {
@@ -518,8 +609,9 @@ static inline void key_generic_cmd(void)
 
 	if (!valid) {
 		asprintf(&error_msg, _("No such command: %s"), cmd);
-		warnbox(error_msg);
+		status_mesg(error_msg, "");
 		mem_free(error_msg);
+		keys_wgetch(win[KEY].p);
 	}
 
 cleanup:
@@ -543,15 +635,6 @@ int main(int argc, char **argv)
 	textdomain(PACKAGE);
 #endif /* ENABLE_NLS */
 
-	/* Thread-safe data structure init */
-	apoint_llist_init();
-	recur_apoint_llist_init();
-
-	/* Initialize non-thread-safe data structures. */
-	event_llist_init();
-	recur_event_llist_init();
-	todo_init_list();
-
 	/*
 	 * Begin by parsing and handling command line arguments.
 	 * The data path is also initialized here.
@@ -565,6 +648,9 @@ int main(int argc, char **argv)
 		io_set_lock();
 	}
 
+	/* System message queue. */
+	que_init();
+
 	/* Begin of interactive mode with ncurses interface. */
 	sigs_init();		/* signal handling init */
 	initscr();		/* start the curses mode */
@@ -619,8 +705,7 @@ int main(int argc, char **argv)
 	config_load();
 	wins_erase_status_bar();
 	io_load_keys(conf.pager);
-	io_load_data(NULL);
-	io_unset_modified();
+	io_load_data(NULL, FORCE);
 	wins_slctd_set(conf.default_panel);
 	wins_resize();
 	/*
@@ -649,9 +734,23 @@ int main(int argc, char **argv)
 	for (;;) {
 		int key;
 
+		while (que_ued()) {
+			que_show();
+			if (conf.systemevents) {
+				que_save();
+				do_storage(0);
+			}
+			wins_update(FLAG_ALL);
+			que_rem();
+		}
+
 		if (resize) {
 			resize = 0;
 			wins_reset();
+			if (conf.multiple_days) {
+				do_storage(0);
+				wins_update(FLAG_APP);
+			}
 		}
 
 		if (want_reload) {
@@ -659,7 +758,10 @@ int main(int argc, char **argv)
 			key_generic_reload();
 		}
 
+		/* Check input loop once every minute. */
+		wtimeout(win[KEY].p, 60000);
 		key = keys_get(win[KEY].p, &count, &reg);
+		wtimeout(win[KEY].p, -1);
 		switch (key) {
 		HANDLE_KEY(KEY_GENERIC_CHANGE_VIEW, key_generic_change_view);
 		HANDLE_KEY(KEY_GENERIC_OTHER_CMD, key_generic_other_cmd);
