@@ -45,6 +45,88 @@
 static vector_t day_items;
 static unsigned day_items_nb = 0;
 
+struct day_item empty_day = { 0, 0, 0, {NULL}};
+
+/*
+ * The day vector, day_items, is continuously rebuilt for display as the
+ * selected day changes, items are added, edited or deleted etc. To keep track
+ * of the selected item of the day vector across rebuilds, data are saved in
+ * here that may later be used to refind the item in the rebuilt day
+ * vector.
+ */
+static struct day_item sel_data = { 0, 0, 0, {NULL}};
+
+/*
+ * Save the item to become the selected APP item.
+ * Public function used to override the setting in do_storage().
+ */
+int day_set_sel_data(struct day_item *d)
+{
+	if (!d)
+		return 0;
+
+	sel_data = *d;
+	return 1;
+}
+
+/*
+ * Return selection data if available.
+ */
+int day_check_sel_data()
+{
+	return (sel_data.order || sel_data.item.apt) ? 1 : 0;
+}
+
+/*
+ * Return the position of the saved selection in the day vector.
+ */
+int day_sel_index(void)
+{
+	int i = 0;
+	struct day_item *p;
+
+	VECTOR_FOREACH(&day_items, i) {
+		p = VECTOR_NTH(&day_items, i);
+		if (p->order == sel_data.order &&
+		    p->item.apt == sel_data.item.apt)
+			return i;
+	}
+	/* Needed as long as ui_day_item_paste() does not set order. */
+	VECTOR_FOREACH(&day_items, i) {
+		p = VECTOR_NTH(&day_items, i);
+		if (p->item.apt == sel_data.item.apt)
+			return i;
+	}
+	/* If still not found, stay on the same day. */
+	VECTOR_FOREACH(&day_items, i) {
+		p = VECTOR_NTH(&day_items, i);
+		if (p->order == update_time_in_date(sel_data.order, 0, 0))
+			return i;
+	}
+	return -1;
+}
+
+/*
+ * Return the number of days to load in the appointments panel.
+ */
+int day_get_days(void)
+{
+	int panel, day;
+
+	if (!conf.multiple_days)
+		return 1;
+
+	panel = win[APP].h - (conf.compact_panels ? 2 : 4);
+	/* Assume one event per day (no event separator). */
+	day = 2 + conf.header_line + conf.day_separator + conf.empty_appt_line;
+
+	if (panel <= day)
+		return 2;
+
+	/* Round up. */
+	return panel / day + (panel % day ? 1 : 0);
+}
+
 static void day_free(struct day_item *day)
 {
 	mem_free(day);
@@ -52,6 +134,7 @@ static void day_free(struct day_item *day)
 
 static void day_init_vector(void)
 {
+	day_items_nb = 0;
 	VECTOR_INIT(&day_items, 16);
 }
 
@@ -72,6 +155,11 @@ static int day_cmp(struct day_item **pa, struct day_item **pb)
 	struct day_item *b = *pb;
 	int a_state, b_state;
 
+	if (a->order < b->order)
+		return -1;
+	if (a->order > b->order)
+		return 1;
+
 	if ((a->type == APPT || a->type == RECUR_APPT) &&
 	    (b->type == APPT || b->type == RECUR_APPT)) {
 		if (a->start < b->start)
@@ -96,11 +184,12 @@ static int day_cmp(struct day_item **pa, struct day_item **pb)
 }
 
 /* Add an item to the current day list. */
-static void day_add_item(int type, long start, union aptev_ptr item)
+static void day_add_item(int type, time_t start, time_t order, union aptev_ptr item)
 {
 	struct day_item *day = mem_malloc(sizeof(struct day_item));
 	day->type = type;
 	day->start = start;
+	day->order = order;
 	day->item = item;
 
 	VECTOR_ADD(&day_items, day);
@@ -188,7 +277,7 @@ int day_item_get_state(struct day_item *day)
 }
 
 /* Add an exception to an item. */
-void day_item_add_exc(struct day_item *day, long date)
+void day_item_add_exc(struct day_item *day, time_t date)
 {
 	switch (day->type) {
 	case RECUR_EVNT:
@@ -207,6 +296,7 @@ void day_item_fork(struct day_item *day_in, struct day_item *day_out)
 {
 	day_out->type = day_in->type;
 	day_out->start = day_in->start;
+	day_out->order = day_in->order;
 
 	switch (day_in->type) {
 	case APPT:
@@ -234,7 +324,7 @@ void day_item_fork(struct day_item *day_in, struct day_item *day_out)
  * dedicated to the selected day.
  * Returns the number of events for the selected day.
  */
-static int day_store_events(long date)
+static int day_store_events(time_t date)
 {
 	llist_item_t *i;
 	union aptev_ptr p;
@@ -244,7 +334,7 @@ static int day_store_events(long date)
 		struct event *ev = LLIST_TS_GET_DATA(i);
 
 		p.ev = ev;
-		day_add_item(EVNT, ev->day, p);
+		day_add_item(EVNT, ev->day, ev->day, p);
 		e_nb++;
 	}
 
@@ -258,7 +348,7 @@ static int day_store_events(long date)
  * dedicated to the selected day.
  * Returns the number of recurrent events for the selected day.
  */
-static int day_store_recur_events(long date)
+static int day_store_recur_events(time_t date)
 {
 	llist_item_t *i;
 	union aptev_ptr p;
@@ -268,8 +358,11 @@ static int day_store_recur_events(long date)
 		struct recur_event *rev = LLIST_TS_GET_DATA(i);
 
 		p.rev = rev;
-		day_add_item(RECUR_EVNT, rev->day, p);
-		e_nb++;
+		time_t occurrence;
+		if (recur_event_find_occurrence(rev, date, &occurrence)) {
+			day_add_item(RECUR_EVNT, rev->day, occurrence, p);
+			e_nb++;
+		}
 	}
 
 	return e_nb;
@@ -282,7 +375,7 @@ static int day_store_recur_events(long date)
  * structure dedicated to the selected day.
  * Returns the number of appointments for the selected day.
  */
-static int day_store_apoints(long date)
+static int day_store_apoints(time_t date)
 {
 	llist_item_t *i;
 	union aptev_ptr p;
@@ -293,11 +386,14 @@ static int day_store_apoints(long date)
 		struct apoint *apt = LLIST_TS_GET_DATA(i);
 
 		p.apt = apt;
-
-		if (apt->start >= date + DAYINSEC)
-			break;
-
-		day_add_item(APPT, apt->start, p);
+		/*
+		 * For appointments continuing from the previous day, order is
+		 * set to midnight to sort it before appointments of the day.
+		 */
+		day_add_item(APPT,
+			     apt->start,
+			     apt->start < date ? date : apt->start,
+			     p);
 		a_nb++;
 	}
 	LLIST_TS_UNLOCK(&alist_p);
@@ -312,10 +408,11 @@ static int day_store_apoints(long date)
  * structure dedicated to the selected day.
  * Returns the number of recurrent appointments for the selected day.
  */
-static int day_store_recur_apoints(long date)
+static int day_store_recur_apoints(time_t date)
 {
 	llist_item_t *i;
 	union aptev_ptr p;
+	time_t occurrence;
 	int a_nb = 0;
 
 	LLIST_TS_LOCK(&recur_alist_p);
@@ -323,10 +420,12 @@ static int day_store_recur_apoints(long date)
 		struct recur_apoint *rapt = LLIST_TS_GET_DATA(i);
 
 		p.rapt = rapt;
-
-		time_t real_start;
-		if (recur_apoint_find_occurrence(rapt, date, &real_start)) {
-			day_add_item(RECUR_APPT, real_start, p);
+		/* As for appointments */
+		if (recur_apoint_find_occurrence(rapt, date, &occurrence)) {
+			day_add_item(RECUR_APPT,
+				     occurrence,
+				     occurrence < date ? date : occurrence,
+				     p);
 			a_nb++;
 		}
 	}
@@ -336,57 +435,56 @@ static int day_store_recur_apoints(long date)
 }
 
 /*
- * Store all of the items to be displayed for the selected day.
- * Items are of four types: recursive events, normal events,
+ * Store all of the items to be displayed for the selected day and the following
+ * (n - 1) days. Items are of four types: recursive events, normal events,
  * recursive appointments and normal appointments.
- * The items are stored in the linked list pointed by day_items
- * and the length of the new pad to write is returned.
- * The number of events and appointments in the current day are also updated.
+ * The items are stored in the day_items vector; the number of events and
+ * appointments in the vector is stored in day_items_nb,
  */
 void
-day_store_items(long date, int include_captions)
+day_store_items(time_t date, int include_captions, int n)
 {
 	unsigned apts, events;
-	union aptev_ptr p = { NULL };
+	union aptev_ptr p = { NULL }, d;
+	int i;
 
 	day_free_vector();
 	day_init_vector();
 
-	if (include_captions)
-		day_add_item(DAY_HEADING, 0, p);
+	for (i = 0; i < n; i++, date = NEXTDAY(date)) {
+		if (YEAR1902_2037 && !check_sec(&date))
+			break;
 
-	events = day_store_recur_events(date);
-	events += day_store_events(date);
-	apts = day_store_recur_apoints(date);
-	apts += day_store_apoints(date);
+		if (include_captions)
+			day_add_item(DAY_HEADING, 0, date, p);
 
-	if (include_captions && events > 0 && apts > 0)
-		day_add_item(DAY_SEPARATOR, 0, p);
+		events = day_store_recur_events(date);
+		events += day_store_events(date);
+		apts = day_store_recur_apoints(date);
+		apts += day_store_apoints(date);
+
+		if (include_captions && events > 0 && apts > 0)
+			day_add_item(EVNT_SEPARATOR, 0, date, p);
+
+		day_items_nb += events + apts;
+
+		if (include_captions && events == 0 && apts == 0) {
+			/* Insert dummy event. */
+			d.ev = &dummy;
+			dummy.mesg = _("(none)");
+			day_add_item(EVNT, DUMMY, date, d);
+			day_items_nb++;
+		}
+
+		if (include_captions) {
+			/* Empty line at end of day if appointments have one. */
+			if (apts == 0 && conf.empty_appt_line)
+				day_add_item(EMPTY_SEPARATOR, 0, ENDOFDAY(date), p);
+			day_add_item(END_SEPARATOR, 0, ENDOFDAY(date), p);
+		}
+	}
 
 	VECTOR_SORT(&day_items, day_cmp);
-	day_items_nb = events + apts;
-}
-
-/*
- * Store the events and appointments for the selected day, and write
- * those items in a pad. If selected day is null, then store items for current
- * day. This is useful to speed up the appointment panel update.
- */
-void day_process_storage(struct date *slctd_date, unsigned day_changed)
-{
-	struct date day;
-
-	if (slctd_date)
-		day = *slctd_date;
-	else
-		ui_calendar_store_current_date(&day);
-
-	/* Inits */
-	if (apad.length != 0)
-		delwin(apad.ptrwin);
-
-	/* Store the events and appointments (recursive and normal items). */
-	day_store_items(date2sec(day, 0, 0), 1);
 }
 
 /*
@@ -394,7 +492,7 @@ void day_process_storage(struct date *slctd_date, unsigned day_changed)
  */
 void
 day_display_item_date(struct day_item *day, WINDOW *win, int incolor,
-		      long date, int y, int x)
+		      time_t date, int y, int x)
 {
 	char a_st[100], a_end[100];
 	char ch_recur, ch_notify;
@@ -448,7 +546,7 @@ day_display_item(struct day_item *day, WINDOW *win, int incolor, int width,
 }
 
 /* Write the appointments and events for the selected day to stdout. */
-void day_write_stdout(long date, const char *fmt_apt, const char *fmt_rapt,
+void day_write_stdout(time_t date, const char *fmt_apt, const char *fmt_rapt,
 		      const char *fmt_ev, const char *fmt_rev, int *limit)
 {
 	int i;
@@ -494,9 +592,7 @@ void day_popup_item(struct day_item *day)
 		struct apoint apt_tmp;
 		apt_tmp.start = day->start;
 		apt_tmp.dur = day_item_get_duration(day);
-		apoint_sec2str(&apt_tmp, ui_calendar_get_slctd_day_sec(),
-			       a_st, a_end);
-
+		apoint_sec2str(&apt_tmp, ui_day_sel_date(), a_st, a_end);
 		item_in_popup(a_st, a_end, day_item_get_mesg(day),
 			      _("Appointment:"));
 	} else {
@@ -506,10 +602,10 @@ void day_popup_item(struct day_item *day)
 }
 
 /*
- * Check whether there is an item on a given day.
- *
- * Returns 2 if the selected day contains a regular event or appointment.
- * Returns 1 if the selected day does not contain a regular event or
+ * Check whether there is an item on a given day and return the colour
+ * attribute for the item:
+ * ATTR_TRUE if the selected day contains a regular event or appointment,
+ * ATTR_LOW if the selected day does not contain a regular event or
  * appointment but an occurrence of a recurrent item. Returns 0 otherwise.
  */
 int day_check_if_item(struct date day)
@@ -517,23 +613,23 @@ int day_check_if_item(struct date day)
 	const time_t t = date2sec(day, 0, 0);
 
 	if (LLIST_FIND_FIRST(&eventlist, (time_t *)&t, event_inday))
-		return 2;
+		return ATTR_TRUE;
 
 	LLIST_TS_LOCK(&alist_p);
 	if (LLIST_TS_FIND_FIRST(&alist_p, (time_t *)&t, apoint_inday)) {
 		LLIST_TS_UNLOCK(&alist_p);
-		return 2;
+		return ATTR_TRUE;
 	}
 	LLIST_TS_UNLOCK(&alist_p);
 
 	if (LLIST_FIND_FIRST(&recur_elist, (time_t *)&t, recur_event_inday))
-		return 1;
+		return ATTR_LOW;
 
 	LLIST_TS_LOCK(&recur_alist_p);
 	if (LLIST_TS_FIND_FIRST(&recur_alist_p, (time_t *)&t,
 				recur_apoint_inday)) {
 		LLIST_TS_UNLOCK(&recur_alist_p);
-		return 1;
+		return ATTR_LOW;
 	}
 	LLIST_TS_UNLOCK(&recur_alist_p);
 
@@ -576,7 +672,7 @@ unsigned day_chk_busy_slices(struct date day, int slicesno, int *slices)
 			      recur_apoint_inday, i) {
 		struct recur_apoint *rapt = LLIST_TS_GET_DATA(i);
 		time_t occurrence;
-		long start, end;
+		time_t start, end;
 
 		if (!recur_apoint_find_occurrence(rapt, t, &occurrence))
 			continue;
@@ -609,8 +705,8 @@ unsigned day_chk_busy_slices(struct date day, int slicesno, int *slices)
 	LLIST_TS_LOCK(&alist_p);
 	LLIST_TS_FIND_FOREACH(&alist_p, (time_t *)&t, apoint_inday, i) {
 		struct apoint *apt = LLIST_TS_GET_DATA(i);
-		long start = get_item_time(apt->start);
-		long end = get_item_time(apt->start + apt->dur);
+		time_t start = get_item_time(apt->start);
+		time_t end = get_item_time(apt->start + apt->dur);
 
 		if (apt->start >= t + DAYINSEC)
 			break;
@@ -639,7 +735,7 @@ unsigned day_chk_busy_slices(struct date day, int slicesno, int *slices)
 }
 
 /* Cut an item so it can be pasted somewhere else later. */
-struct day_item *day_cut_item(long date, int item_number)
+struct day_item *day_cut_item(int item_number)
 {
 	struct day_item *p = day_get_item(item_number);
 
@@ -665,25 +761,52 @@ struct day_item *day_cut_item(long date, int item_number)
 }
 
 /* Paste a previously cut item. */
-int day_paste_item(struct day_item *p, long date)
+int day_paste_item(struct day_item *p, time_t date)
 {
 	if (!p->type) {
 		/* No previously cut item. */
 		return 0;
 	}
+	/*
+	 * Valid until date of recurrent items?
+	 * Careful: p->start is not yet set.
+	 */
+	time_t until;
 
 	switch (p->type) {
 	case EVNT:
 		event_paste_item(p->item.ev, date);
 		break;
 	case RECUR_EVNT:
-		recur_event_paste_item(p->item.rev, date);
+		/* want: until = shift + old_until */
+		if (p->item.rev->rpt->until &&
+		    overflow_add(
+			date - p->item.rev->day,
+			p->item.rev->rpt->until,
+			&until)
+		)
+			return 0;
+		if (check_sec(&until))
+			recur_event_paste_item(p->item.rev, date);
+		else
+			return 0;
 		break;
 	case APPT:
 		apoint_paste_item(p->item.apt, date);
 		break;
 	case RECUR_APPT:
-		recur_apoint_paste_item(p->item.rapt, date);
+		/* wanted: until = shift + old_until */
+		if (p->item.rapt->rpt->until &&
+		    overflow_add(
+			date - update_time_in_date(p->item.rapt->start, 0, 0),
+			p->item.rapt->rpt->until,
+			&until)
+		)
+			return 0;
+		if (check_sec(&until))
+			recur_apoint_paste_item(p->item.rapt, date);
+		else
+			return 0;
 		break;
 	default:
 		EXIT(_("unknown item type"));
@@ -693,26 +816,6 @@ int day_paste_item(struct day_item *p, long date)
 	return p->type;
 }
 
-/* Returns the position corresponding to a given item. */
-int day_get_position_by_aptev_ptr(union aptev_ptr aptevp)
-{
-	int n = 0;
-
-	VECTOR_FOREACH(&day_items, n) {
-		struct day_item *p = VECTOR_NTH(&day_items, n);
-		/* Compare pointers. */
-		if (p->item.ev == aptevp.ev)
-			return n;
-	}
-
-	return -1;
-}
-
-int day_get_position(struct day_item *needle)
-{
-	return day_get_position_by_aptev_ptr(needle->item);
-}
-
 /* Returns a structure containing the selected item. */
 struct day_item *day_get_item(int item_number)
 {
